/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：C++ 信息安全性设计准则
 * 项目描述：用于检查C++源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package com.keyware.sonar.cxx.rules.checkers;

import com.keyware.sonar.cxx.SubscriptionAstVisitor;
import com.sonar.cxx.sslr.api.AstNode;
import com.sonar.cxx.sslr.api.AstNodeType;
import com.sonar.cxx.sslr.api.Grammar;
import org.sonar.check.Priority;
import org.sonar.check.Rule;
import org.sonar.cxx.parser.CxxGrammarImpl;
import org.sonar.cxx.squidbridge.annotations.ActivatedByDefault;
import org.sonar.cxx.squidbridge.annotations.SqaleConstantRemediation;
import org.sonar.cxx.squidbridge.checks.SquidCheck;

import java.util.ArrayList;
import java.util.List;

/**
 * TODO PassWordCountChecker
 *
 * @author RenFengJiang
 * @date 2024/1/13
 */
@Rule(key = "PassWordCountChecker", name = "使用盐值计算散列值", description = "使用盐值计算散列值，增加攻击者破解口令的难度", priority = Priority.INFO, tags = {"28suo"})
@ActivatedByDefault
@SqaleConstantRemediation("5min")
public class PassWordCountChecker extends SquidCheck<Grammar> {

    @Override
    public void init() {
        // 订阅要检查AST节点类型，用于在visitNode方法中检查该类型节点
        this.subscribeTo(
                CxxGrammarImpl.functionBody
        );
    }


    /**
     * 检查AST节点
     *
     * @param astNode 要处理的AST节点，该节点类型为通过subscribeTo方法订阅的类型
     */
    @Override
    public void visitNode(AstNode astNode) {
        BodyWay bodyWay = new BodyWay(this);
        bodyWay.accept(astNode);
    }

    class BodyWay extends SubscriptionAstVisitor{

        /**
         * 构造函数需要传入初代访问器
         *
         * @param checker 初代规则检查器
         */
        public BodyWay(SquidCheck<Grammar> checker) {
            super(checker);
        }

        @Override
        public List<AstNodeType> visitNodeTypes() {
            return List.of(CxxGrammarImpl.functionBody);
        }

        @Override
        public void visitNode(AstNode astNode) {
            //获取方法参数列表
            List<AstNode> nodeDescendants = astNode.getDescendants(CxxGrammarImpl.expressionList);
            List<String> seqLists = new ArrayList<>();
            for(AstNode desc :nodeDescendants){
                seqLists.add(desc.getTokenValue());
            }
            //判读入参有没有进行赋值操作
            List<AstNode> astNodes = astNode.getDescendants(CxxGrammarImpl.assignmentExpression);
            List<String> aslists = new ArrayList<>();
            for(AstNode ast : astNodes){
                if(seqLists.contains(ast.getTokenValue())){
                    aslists.add(ast.getTokenValue());
                }
            }
            //获取到声明hash变量名
            List<String> lists = new ArrayList<>();
            List<AstNode> astNodeDescendants = astNode.getDescendants(CxxGrammarImpl.simpleDeclaration);
            for(AstNode des :astNodeDescendants){
                if("std".equals(des.getTokenValue())){
                    List<AstNode> descendants = des.getDescendants(CxxGrammarImpl.typeName);
                    for (AstNode dan : descendants){
                        if("hash".equals(dan.getTokenValue())){
                            List<AstNode> nodeList = des.getDescendants(CxxGrammarImpl.initDeclarator);
                            for (AstNode lis:nodeList) {
                                lists.add(lis.getTokenValue());
                            }
                        }
                    }
                }
            }
            //判断生成的散列值参数是否是进行赋后的入参
            List<AstNode> descendants = astNode.getDescendants(CxxGrammarImpl.postfixExpression);
            for (AstNode desc:descendants) {
                if(lists.contains(desc.getTokenValue())){
                    List<AstNode> descDescendants = desc.getDescendants(CxxGrammarImpl.expressionList);
                    for(AstNode dd : descDescendants){
                        if(!aslists.contains(dd.getTokenValue())){
                            reportIssue(dd, "使用盐值计算散列值");
                        }
                    }
                }
            }
        }
    }
}
